@@ -35,13 +35,14 @@ handle_add_notification_gtk (XdpImplNotification *object,
                              const char *arg_id,
                              GVariant *arg_notification)
 {
-  org_gtk_notifications_call_add_notification (gtk_notifications,
-                                               arg_app_id,
-                                               arg_id,
-                                               arg_notification,
-                                               NULL,
-                                               NULL,
-                                               NULL);
+  if (gtk_notifications)
+    org_gtk_notifications_call_add_notification (gtk_notifications,
+                                                 arg_app_id,
+                                                 arg_id,
+                                                 arg_notification,
+                                                 NULL,
+                                                 NULL,
+                                                 NULL);
 
   xdp_impl_notification_complete_add_notification (object, invocation);
 
@@ -54,12 +55,13 @@ handle_remove_notification_gtk (XdpImplNotification *object,
                                 const char *arg_app_id,
                                 const char *arg_id)
 {
-  org_gtk_notifications_call_remove_notification (gtk_notifications,
-                                                  arg_app_id,
-                                                  arg_id,
-                                                  NULL,
-                                                  NULL,
-                                                  NULL);
+  if (gtk_notifications)
+    org_gtk_notifications_call_remove_notification (gtk_notifications,
+                                                    arg_app_id,
+                                                    arg_id,
+                                                    NULL,
+                                                    NULL,
+                                                    NULL);
 
   xdp_impl_notification_complete_remove_notification (object, invocation);
 
@@ -149,6 +151,7 @@ app_path_for_id (const gchar *app_id)
 static void
 activate_action (GDBusConnection *connection,
                  const char *app_id,
+                 const char *id,
                  const char *name,
                  GVariant *parameter)
 {
@@ -175,6 +178,9 @@ activate_action (GDBusConnection *connection,
     }
   else
     {
+      g_autoptr(GVariant) ret = NULL;
+      GVariantBuilder parms;
+
       g_dbus_connection_call (connection,
                               app_id,
                               object_path,
@@ -185,6 +191,20 @@ activate_action (GDBusConnection *connection,
                               NULL,
                               G_DBUS_CALL_FLAGS_NONE,
                               -1, NULL, NULL, NULL);
+
+      g_variant_builder_init (&parms, G_VARIANT_TYPE ("av"));
+      if (parameter)
+        g_variant_builder_add (&parms, "v", parameter);
+
+      g_dbus_connection_emit_signal (connection,
+                                     NULL,
+                                     "/org/freedesktop/portal/desktop",
+                                     "org.freedesktop.impl.portal.Notification",
+                                     "ActionInvoked",
+                                     g_variant_new ("(sss@av)",
+                                                    app_id, id, name,
+                                                    g_variant_builder_end (&parms)),
+                                     NULL);
     }
 }
 
@@ -224,6 +244,7 @@ notify_signal (GDBusConnection *connection,
         {
           activate_action (connection,
                            n->app_id,
+                           n->id,
                            n->default_action,
                            n->default_action_target);
         }
@@ -236,6 +257,7 @@ notify_signal (GDBusConnection *connection,
             {
               activate_action (connection,
                                n->app_id,
+                               n->id,
                                name,
                                target);
               g_free (name);
@@ -482,31 +504,79 @@ handle_remove_notification_fdo (XdpImplNotification *object,
     }
 }
 
+static gboolean
+has_unprefixed_action (GVariant *notification)
+{
+  const char *action;
+  g_autoptr(GVariant) buttons = NULL;
+  int i;
+
+  if (g_variant_lookup (notification, "default-action", "&s", &action) &&
+      !g_str_has_prefix (action, "app."))
+    return TRUE;
+
+  buttons = g_variant_lookup_value (notification, "buttons", G_VARIANT_TYPE("aa{sv}"));
+  if (buttons)
+    for (i = 0; i < g_variant_n_children (buttons); i++)
+      {
+        g_autoptr(GVariant) button = NULL;
+
+        button = g_variant_get_child_value (buttons, i);
+        if (g_variant_lookup (button, "action", "&s", &action) &&
+            !g_str_has_prefix (action, "app."))
+          return TRUE;
+      }
+
+  return FALSE;
+}
+
+static void
+handle_add_notification (XdpImplNotification *object,
+                         GDBusMethodInvocation *invocation,
+                         const gchar *arg_app_id,
+                         const gchar *arg_id,
+                         GVariant *arg_notification)
+{
+  if (gtk_notifications == NULL ||
+      g_dbus_proxy_get_name_owner (G_DBUS_PROXY (gtk_notifications)) == NULL ||
+      has_unprefixed_action (arg_notification))
+    handle_add_notification_fdo (object, invocation, arg_app_id, arg_id, arg_notification);
+  else
+    handle_add_notification_gtk (object, invocation, arg_app_id, arg_id, arg_notification);
+}
+
+static void
+handle_remove_notification (XdpImplNotification *object,
+                            GDBusMethodInvocation *invocation,
+                            const gchar *arg_app_id,
+                            const gchar *arg_id)
+{
+  FdoNotification *n;
+
+  n = fdo_find_notification (arg_app_id, arg_id);
+  if (n)
+    handle_remove_notification_fdo (object, invocation, arg_app_id, arg_id);
+  else
+    handle_remove_notification_gtk (object, invocation, arg_app_id, arg_id);
+}
+
 gboolean
 notification_init (GDBusConnection *bus,
                    GError **error)
 {
   GDBusInterfaceSkeleton *helper;
 
-  helper = G_DBUS_INTERFACE_SKELETON (xdp_impl_notification_skeleton_new ());
-
   gtk_notifications = org_gtk_notifications_proxy_new_sync (bus,
                                                             G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                                                             "org.gtk.Notifications",
                                                             "/org/gtk/Notifications",
                                                             NULL,
                                                             NULL);
-  if (gtk_notifications != NULL &&
-      g_dbus_proxy_get_name_owner (G_DBUS_PROXY (gtk_notifications)) != NULL)
-    {
-      g_signal_connect (helper, "handle-add-notification", G_CALLBACK (handle_add_notification_gtk), NULL);
-      g_signal_connect (helper, "handle-remove-notification", G_CALLBACK (handle_remove_notification_gtk), NULL);
-    }
-  else
-    {
-      g_signal_connect (helper, "handle-add-notification", G_CALLBACK (handle_add_notification_fdo), NULL);
-      g_signal_connect (helper, "handle-remove-notification", G_CALLBACK (handle_remove_notification_fdo), NULL);
-    }
+
+  helper = G_DBUS_INTERFACE_SKELETON (xdp_impl_notification_skeleton_new ());
+
+  g_signal_connect (helper, "handle-add-notification", G_CALLBACK (handle_add_notification), NULL);
+  g_signal_connect (helper, "handle-remove-notification", G_CALLBACK (handle_remove_notification), NULL);
 
   if (!g_dbus_interface_skeleton_export (helper,
                                          bus,
